from pint import Quantity

from .database import select_thermocalc_database
from tc.schema import Composition

T_AXIS_MIN = 500.0
T_AXIS_MAX = 3500.0

def get_temperature_solidus_liquidus(composition: Composition) -> tuple[float, float] 
    """
    Uses Thermo-Calc to compute solidus & liquidus from elements/fractions.
    Returns (solidus_K, liquidus_K, database_name)
    """
    database = select_thermocalc_database(composition)

    print(database)
    return database

    # with TCPython() as tc:
    #     tc.set_cache_folder("cache")
    #     tc.set_ges_version(6)
    #
    #     calc = (
    #         tc.select_database_and_elements(database, elements)
    #           .get_system()
    #           .with_property_diagram_calculation()
    #           .with_axis(
    #               CalculationAxis(ThermodynamicQuantity.temperature())
    #               .set_min(T_axis_min)
    #               .set_max(T_axis_max)
    #               .with_axis_type(Linear().set_min_nr_of_steps(50))
    #           )
    #     )
    #
    #     # Set composition via mass fractions. Skip the first element; Thermo-Calc will normalize.
    #     for el, wf in list(fractions.items())[1:]:
    #         calc.set_condition(f"W({el})", wf)
    #
    #     diagram = calc.calculate()
    #     diagram.set_phase_name_style(PhaseNameStyle.ALL)
    #
    #     groups = diagram.get_values_grouped_by_quantity_of(
    #         ThermodynamicQuantity.temperature(),
    #         ThermodynamicQuantity.volume_fraction_of_a_phase("LIQUID"),
    #     )
    #
    # solidus_T = None
    # liquidus_T = None
    #
    # for group in groups.values():
    #     xT = group.x
    #     yL = group.y
    #     last_zero_idx = max((i for i, y in enumerate(yL) if abs(y) < 1e-12), default=None)
    #     first_one_idx = None
    #     if last_zero_idx is not None:
    #         for i in range(last_zero_idx + 1, len(yL)):
    #             if abs(yL[i] - 1.0) < 1e-12:
    #                 first_one_idx = i
    #                 break
    #     if last_zero_idx is not None:
    #         solidus_T = xT[last_zero_idx]
    #     if first_one_idx is not None:
    #         liquidus_T = xT[first_one_idx]
    #
    # if solidus_T is None or liquidus_T is None:
    #     raise RuntimeError(
    #         f"Could not determine solidus/liquidus using {database} "
    #         f"in range [{T_axis_min:.1f}, {T_axis_max:.1f}] K."
    #     )
    #
    # return float(solidus_T), float(liquidus_T), database




from pydantic import BaseModel


class Composition(BaseModel):

    # Base elements
    Fe: float | None = None
    C: float | None = None

    # TCFe alloys
    H: float | None = None
    Mg: float | None = None
    Ca: float | None = None
    Y: float | None = None
    Ti: float | None = None
    Zr: float | None = None
    V: float | None = None
    Nb: float | None = None
    Ta: float | None = None
    Cr: float | None = None
    Mo: float | None = None
    W: float | None = None
    Mn: float | None = None
    Ru: float | None = None
    Co: float | None = None
    Ni: float | None = None
    Cu: float | None = None
    Zn: float | None = None
    B: float | None = None
    Al: float | None = None
    Si: float | None = None
    N: float | None = None
    P: float | None = None
    O: float | None = None
    S: float | None = None
    Ar: float | None = None
    Ce: float | None = None

    # TCNi12 extra elements
    Hf: float | None = None
    Re: float | None = None
    Pd: float | None = None
    Pt: float | None = None

    # TCAl9 extra elements
    Li: float | None = None
    Na: float | None = None
    K: float | None = None
    Be: float | None = None
    Sr: float | None = None
    Ba: float | None = None
    Sc: float | None = None
    Ga: float | None = None
    Ge: float | None = None
    Ag: float | None = None
    In: float | None = None
    Cd: float | None = None
    Sn: float | None = None
    Sb: float | None = None
    Te: float | None = None
    Se: float | None = None
    Pb: float | None = None
    Bi: float | None = None
    La: float | None = None
    Pr: float | None = None
    Nd: float | None = None
    Er: float | None = None

    # TCHEA7
    Ir: float | None = None



from .alloy_properties import alloy_properties


def alloy_names() -> list[str]:
    comp_rows = alloy_properties()
    names: list[str] = []
    for row in comp_rows:
        name = row.get("Name")
        if name is not None:
            names.append(name)

    return names



from .alloy_list import register_alloy_list
from .alloy_composition import register_alloy_composition
from .temperature import register_alloy_temperature

__all__ = ["register_alloy_list", "register_alloy_composition", "register_alloy_temperature"]



from mcp.server import FastMCP


def register_alloy_temperature(app: FastMCP):
    from tc.mcp.types import ToolSuccess, ToolError
    from tc.mcp.utils import tool_success, tool_error
    from tc.schema.composition import Composition

    @app.tool(
        title="Get Solidus Liquidus temperature",
        description="Calculate the solidus and liquids temperature given an alloy composition.",
        structured_output=True,
    )
    def temperature_solidus_liquidus(composition: Composition) -> ToolSuccess[tuple[float, float]] | ToolError:
        from tc.alloy.temperature import get_temperature_solidus_liquidus

        try:
            return tool_success(get_temperature_solidus_liquidus(composition))

        except Exception as e:
            return tool_error(
                "Failed to get alloy liquidus and solidus temperatures",
                "ALLOY_TEMPERATURE_FAILED",
                exception_type=type(e).__name__,
                exception_message=str(e),
            )

    _ = temperature_solidus_liquidus



from .__main__ import app

from .alloy_composition import register_alloy_composition
from .alloy_list import register_alloy_list

# from .initialize_power_velocity_range import (
#     register_process_map_initialize_power_velocity_range,
# )
#
# from .generate_process_map import register_process_map_generate_melt_pool_measurements

# _ = register_process_map_initialize_power_velocity_range(app)
# _ = register_process_map_generate_melt_pool_measurements(app)
_ = register_alloy_composition(app)
_ = register_alloy_list(app)

__all__ = ["app"]



from mcp.server import FastMCP


def register_alloy_composition(app: FastMCP):
    from tc.mcp.types import ToolSuccess, ToolError
    from tc.mcp.utils import tool_success, tool_error
    from tc.alloy.types import Alloy
    from tc.schema.composition import Composition

    @app.tool(
        title="Get Alloy Composition",
        description="Provide the elements and composition for an alloy.",
        structured_output=True,
    )
    def alloy_composition(alloy: Alloy) -> ToolSuccess[Composition | None] | ToolError:
        from tc.alloy.alloy_composition import get_alloy_composition

        try:
            return tool_success(get_alloy_composition(alloy))

        except Exception as e:
            return tool_error(
                "Failed to get alloy composition",
                "ALLOY_COMPOSITION_FAILED",
                exception_type=type(e).__name__,
                exception_message=str(e),
            )

    _ = alloy_composition



from .alloy_compositions import alloy_compositions
from .types import Alloy

from tc.schema.composition import Composition


def get_alloy_composition(alloy: Alloy) -> Composition:
    """
    Returns (element_list, weight_fractions_dict) for a known alloy, normalized to sum=1.0.
    """
    compositions = alloy_compositions()
    composition = compositions.get(alloy.value)

    if composition is not None:

        out: dict[str, float] = {}
        for k, v in composition.items():
            if k in ("", None):
                continue
            out[k] = float(v)

        out = {k: v for k, v in out.items() if v and v > 0.0}
        total = sum(out.values())

        # If looks like percent, normalize from 100; otherwise from sum
        if abs(total - 100.0) < 1e-6:
            fractions = {el: wt / 100.0 for el, wt in out.items()}
        else:
            fractions = {el: wt / total for el, wt in out.items()}
        fractions = dict(sorted(fractions.items(), key=lambda kv: kv[1], reverse=True))
        return Composition(**fractions)

    else:
        raise ValueError(f"No numeric composition fields found for '{alloy}'.")



from mcp.server import FastMCP


def register_alloy_list(app: FastMCP):
    from tc.mcp.types import ToolSuccess
    from tc.mcp.utils import tool_success

    @app.tool(
        title="List Alloys",
        description="Provides a list known alloys that have compositions and material properties.",
        structured_output=True,
    )
    def alloy_list() -> ToolSuccess[list[str] | None]:
        from tc.alloy.alloy_list import alloy_names

        return tool_success(alloy_names())

    _ = alloy_list



def alloy_properties() -> list[dict[str, str]]:
    return [
        {
            "Name": "Ti6Al4V",
            "Density": "4.43",
            "Thermal conductivity": "7.1",
            "Melting temperature (C)": "1695",
            "Coefficient of thermal expansion": "8.6",
            "Specific heat": "561.5",
        },
        {
            "Name": "IN718",
            "Density": "8.19",
            "Thermal conductivity": "9",
            "Melting temperature (C)": "1298",
            "Coefficient of thermal expansion": "14",
            "Specific heat": "435",
        },
        {
            "Name": "SS316L",
            "Density": "7.99",
            "Thermal conductivity": "16.2",
            "Melting temperature (C)": "1385",
            "Coefficient of thermal expansion": "16",
            "Specific heat": "500",
        },
        {
            "Name": "AlSi10Mg",
            "Density": "2.68",
            "Thermal conductivity": "160",
            "Melting temperature (C)": "580",
            "Coefficient of thermal expansion": "20.5",
            "Specific heat": "910",
        },
        {
            "Name": "IN625",
            "Density": "8.44",
            "Thermal conductivity": "9.2",
            "Melting temperature (C)": "1320",
            "Coefficient of thermal expansion": "12.8",
            "Specific heat": "429",
        },
        {
            "Name": "Maraging steel M300",
            "Density": "8.1",
            "Thermal conductivity": "14.2",
            "Melting temperature (C)": "1413",
            "Coefficient of thermal expansion": "10.3",
            "Specific heat": "452",
        },
        {
            "Name": "Ti6Al4V ELI",
            "Density": "4.42",
            "Thermal conductivity": "7",
            "Melting temperature (C)": "1650",
            "Coefficient of thermal expansion": "8.5",
            "Specific heat": "526",
        },
        {
            "Name": "SS17-4PH",
            "Density": "7.8",
            "Thermal conductivity": "16",
            "Melting temperature (C)": "1420",
            "Coefficient of thermal expansion": "10.4",
            "Specific heat": "460",
        },
        {
            "Name": "CoCrMo",
            "Density": "8.4",
            "Thermal conductivity": "13",
            "Melting temperature (C)": "1390",
            "Coefficient of thermal expansion": "14",
            "Specific heat": "450",
        },
        {
            "Name": "H13 Tool Steel",
            "Density": "7.8",
            "Thermal conductivity": "28.6",
            "Melting temperature (C)": "1427",
            "Coefficient of thermal expansion": "10.4",
            "Specific heat": "460",
        },
        {
            "Name": "AlSi7Mg0.6",
            "Density": "2.67",
            "Thermal conductivity": "140",
            "Melting temperature (C)": "570",
            "Coefficient of thermal expansion": "21.5",
            "Specific heat": "890",
        },
        {
            "Name": "SS15-5PH",
            "Density": "7.8",
            "Thermal conductivity": "11",
            "Melting temperature (C)": "1420",
            "Coefficient of thermal expansion": "10.8",
            "Specific heat": "420",
        },
        {
            "Name": "CuCrZr",
            "Density": "8.9",
            "Thermal conductivity": "320",
            "Melting temperature (C)": "1085",
            "Coefficient of thermal expansion": "17",
            "Specific heat": "380",
        },
        {
            "Name": "Al12Si",
            "Density": "2.7",
            "Thermal conductivity": "175",
            "Melting temperature (C)": "680",
            "Coefficient of thermal expansion": "22.3",
            "Specific heat": "922",
        },
        {
            "Name": "Copper",
            "Density": "8.6",
            "Thermal conductivity": "350",
            "Melting temperature (C)": "1085",
            "Coefficient of thermal expansion": "17.01",
            "Specific heat": "435",
        },
        {
            "Name": "Hastelloy X",
            "Density": "8.2",
            "Thermal conductivity": "9.1",
            "Melting temperature (C)": "1355",
            "Coefficient of thermal expansion": "13.9",
            "Specific heat": "486",
        },
        {
            "Name": "CoCr",
            "Density": "8.3",
            "Thermal conductivity": "13",
            "Melting temperature (C)": "1413.75",
            "Coefficient of thermal expansion": "12",
            "Specific heat": "390",
        },
        {
            "Name": "Hastelloy C22",
            "Density": "8.69",
            "Thermal conductivity": "8.9",
            "Melting temperature (C)": "1399",
            "Coefficient of thermal expansion": "12.4",
            "Specific heat": "414",
        },
        {
            "Name": "SS254",
            "Density": "8.07",
            "Thermal conductivity": "14",
            "Melting temperature (C)": "1355",
            "Coefficient of thermal expansion": "14.8",
            "Specific heat": "500",
        },
        {
            "Name": "CuNi2SiCr",
            "Density": "8.84",
            "Thermal conductivity": "90",
            "Melting temperature (C)": "1050",
            "Coefficient of thermal expansion": "17.5",
            "Specific heat": "397",
        },
        {
            "Name": "D2 Tool Steel",
            "Density": "7.7",
            "Thermal conductivity": "20",
            "Melting temperature (C)": "1421",
            "Coefficient of thermal expansion": "10.3",
            "Specific heat": "460",
        },
        {
            "Name": "SS4340",
            "Density": "7.85",
            "Thermal conductivity": "44.5",
            "Melting temperature (C)": "1427",
            "Coefficient of thermal expansion": "12.3",
            "Specific heat": "475",
        },
        {
            "Name": "Invar36",
            "Density": "8.1",
            "Thermal conductivity": "12.8",
            "Melting temperature (C)": "1727",
            "Coefficient of thermal expansion": "1.5",
            "Specific heat": "515",
        },
        {
            "Name": "SS304L",
            "Density": "8",
            "Thermal conductivity": "16.2",
            "Melting temperature (C)": "1693",
            "Coefficient of thermal expansion": "17.3",
            "Specific heat": "500",
        },
        {
            "Name": "Ti grade 2",
            "Density": "4.5",
            "Thermal conductivity": "22",
            "Melting temperature (C)": "1665",
            "Coefficient of thermal expansion": "8.6",
            "Specific heat": "523",
        },
        {
            "Name": "K500",
            "Density": "8.43",
            "Thermal conductivity": "17.5",
            "Melting temperature (C)": "1350",
            "Coefficient of thermal expansion": "13.4",
            "Specific heat": "418",
        },
        {
            "Name": "M2 Tool Steel",
            "Density": "8.055",
            "Thermal conductivity": "41.5",
            "Melting temperature (C)": "1450",
            "Coefficient of thermal expansion": "10",
            "Specific heat": "460",
        },
        {
            "Name": "Ti grade 1",
            "Density": "4.51",
            "Thermal conductivity": "16",
            "Melting temperature (C)": "1668",
            "Coefficient of thermal expansion": "7.17",
            "Specific heat": "520",
        },
        {
            "Name": "CoCr28Mo6",
            "Density": "8.47",
            "Thermal conductivity": "12.5",
            "Melting temperature (C)": "1390",
            "Coefficient of thermal expansion": "13.2",
            "Specific heat": "450",
        },
        {
            "Name": "IN690",
            "Density": "8.19",
            "Thermal conductivity": "13.5",
            "Melting temperature (C)": "1360",
            "Coefficient of thermal expansion": "14.06",
            "Specific heat": "450",
        },
        {
            "Name": "CuCP",
            "Density": "8.9",
            "Thermal conductivity": "320",
            "Melting temperature (C)": "1085",
            "Coefficient of thermal expansion": "17",
            "Specific heat": "385",
        },
        {
            "Name": "TiAlNbMo",
            "Density": "4.112",
            "Thermal conductivity": "22",
            "Melting temperature (C)": "1773",
            "Coefficient of thermal expansion": "12",
            "Specific heat": "600",
        },
        {
            "Name": "SS304",
            "Density": "8",
            "Thermal conductivity": "16.2",
            "Melting temperature (C)": "1693",
            "Coefficient of thermal expansion": "17.3",
            "Specific heat": "500",
        },
        {
            "Name": "TiAl",
            "Density": "4.05",
            "Thermal conductivity": "15",
            "Melting temperature (C)": "1460",
            "Coefficient of thermal expansion": "9",
            "Specific heat": "600",
        },
        {
            "Name": "HY100",
            "Density": "7.75",
            "Thermal conductivity": "34",
            "Melting temperature (C)": "1520",
            "Coefficient of thermal expansion": "14",
            "Specific heat": "460",
        },
        {
            "Name": "AlSi12",
            "Density": "2.685",
            "Thermal conductivity": "140",
            "Melting temperature (C)": "579",
            "Coefficient of thermal expansion": "20",
            "Specific heat": "548.6",
        },
        {
            "Name": "CoCrF75",
            "Density": "8.35",
            "Thermal conductivity": "14",
            "Melting temperature (C)": "1390",
            "Coefficient of thermal expansion": "14",
            "Specific heat": "390",
        },
        {
            "Name": "AZ91D",
            "Density": "1.81",
            "Thermal conductivity": "72",
            "Melting temperature (C)": "533",
            "Coefficient of thermal expansion": "25.2",
            "Specific heat": "1047",
        },
        {
            "Name": "CoCrW",
            "Density": "8.6",
            "Thermal conductivity": "15",
            "Melting temperature (C)": "1370",
            "Coefficient of thermal expansion": "14.1",
            "Specific heat": "390",
        },
        {
            "Name": "SS420C",
            "Density": "7.8",
            "Thermal conductivity": "24.9",
            "Melting temperature (C)": "1480",
            "Coefficient of thermal expansion": "10.3",
            "Specific heat": "460",
        },
        {
            "Name": "SS4130",
            "Density": "7.85",
            "Thermal conductivity": "42.7",
            "Melting temperature (C)": "1432",
            "Coefficient of thermal expansion": "14.6",
            "Specific heat": "523",
        },
        {
            "Name": "CuCrZrTi",
            "Density": "8.9",
            "Thermal conductivity": "320",
            "Melting temperature (C)": "1085",
            "Coefficient of thermal expansion": "17",
            "Specific heat": "380",
        },
        {
            "Name": "8620L Steel",
            "Density": "7.85",
            "Thermal conductivity": "46.6",
            "Melting temperature (C)": "1430",
            "Coefficient of thermal expansion": "6.6",
            "Specific heat": "475",
        },
        {
            "Name": "Al2319",
            "Density": "2.84",
            "Thermal conductivity": "170",
            "Melting temperature (C)": "600",
            "Coefficient of thermal expansion": "22.5",
            "Specific heat": "864",
        },
        {
            "Name": "Iron (0.004%C)",
            "Density": "7.86",
            "Thermal conductivity": "50",
            "Melting temperature (C)": "1535",
            "Coefficient of thermal expansion": "7",
            "Specific heat": "460",
        },
        {
            "Name": "SS316",
            "Density": "7.99",
            "Thermal conductivity": "16.2",
            "Melting temperature (C)": "1385",
            "Coefficient of thermal expansion": "16",
            "Specific heat": "500",
        },
        {
            "Name": "Iron (0.02%C)",
            "Density": "7.86",
            "Thermal conductivity": "50",
            "Melting temperature (C)": "1535",
            "Coefficient of thermal expansion": "7",
            "Specific heat": "460",
        },
        {
            "Name": "Mar M 247",
            "Density": "7.8",
            "Thermal conductivity": "15",
            "Melting temperature (C)": "1220",
            "Coefficient of thermal expansion": "12.8",
            "Specific heat": "230",
        },
        {
            "Name": "W22Fe33Ni",
            "Density": "17.6",
            "Thermal conductivity": "27",
            "Melting temperature (C)": "3410",
            "Coefficient of thermal expansion": "4.5",
            "Specific heat": "134",
        },
        {
            "Name": "4140 Steel",
            "Density": "7.85",
            "Thermal conductivity": "42.6",
            "Melting temperature (C)": "1416",
            "Coefficient of thermal expansion": "12.2",
            "Specific heat": "473",
        },
        {
            "Name": "S50C",
            "Density": "7.85",
            "Thermal conductivity": "21",
            "Melting temperature (C)": "1380",
            "Coefficient of thermal expansion": "13.2",
            "Specific heat": "486",
        },
        {
            "Name": "S75C",
            "Density": "7.85",
            "Thermal conductivity": "25",
            "Melting temperature (C)": "1400",
            "Coefficient of thermal expansion": "13",
            "Specific heat": "480",
        },
        {
            "Name": "A2 Tool Steel",
            "Density": "7.86",
            "Thermal conductivity": "26",
            "Melting temperature (C)": "1420",
            "Coefficient of thermal expansion": "10.6",
            "Specific heat": "461",
        },
        {
            "Name": "S105C",
            "Density": "7.872",
            "Thermal conductivity": "53",
            "Melting temperature (C)": "1430",
            "Coefficient of thermal expansion": "12.6",
            "Specific heat": "481",
        },
        {
            "Name": "SS CX",
            "Density": "7.69",
            "Thermal conductivity": "16.2",
            "Melting temperature (C)": "1450",
            "Coefficient of thermal expansion": "11.1",
            "Specific heat": "502",
        },
        {
            "Name": "7050 alloy",
            "Density": "2.7",
            "Thermal conductivity": "180",
            "Melting temperature (C)": "494",
            "Coefficient of thermal expansion": "23.5",
            "Specific heat": "860",
        },
        {
            "Name": "S33C",
            "Density": "7.85",
            "Thermal conductivity": "34",
            "Melting temperature (C)": "1400",
            "Coefficient of thermal expansion": "13",
            "Specific heat": "475",
        },
        {
            "Name": "IN600",
            "Density": "8.47",
            "Thermal conductivity": "14.8",
            "Melting temperature (C)": "1380",
            "Coefficient of thermal expansion": "13.3",
            "Specific heat": "444",
        },
        {
            "Name": "Nb",
            "Density": "8.4",
            "Thermal conductivity": "53.7",
            "Melting temperature (C)": "2410",
            "Coefficient of thermal expansion": "7.1",
            "Specific heat": "260",
        },
    ]
